"""DSSE: A Multi-Agent Reinforcement Learning Environment for SAR missions using drone swarms."""

from .environment.env import DroneSwarmSearch
from .environment.coverage_env import CoverageDroneSwarmSearch
from .environment.constants import Actions

__all__ = ["DroneSwarmSearch", "CoverageDroneSwarmSearch", "Actions"]



from random import random, randint
import functools
import numpy as np
from gymnasium.spaces import MultiDiscrete, Discrete, Box, Tuple
from .constants import RED, GREEN, Actions, Reward
from .entities.person import Person
from .env_base import DroneSwarmSearchBase
from .simulation.dynamic_probability import ProbabilityMatrix


class DroneSwarmSearch(DroneSwarmSearchBase):
    """
    PettingZoo based environment for SAR missions using drones.
    """

    possible_actions = {action for action in Actions}
    metadata = {
        "name": "DroneSwarmSearchV0",
    }

    reward_scheme = Reward(
        default=0.1,
        leave_grid=-200,
        exceed_timestep=-200,
        drones_collision=-200,
        search_cell=1,
        search_and_find=200,
    )

    def __init__(
        self,
        grid_size=20,
        render_mode="ansi",
        render_grid=False,
        render_gradient=True,
        vector=(1.1, 1),
        dispersion_inc=0.1,
        dispersion_start=0.5,
        timestep_limit=100,
        person_amount=1,
        person_initial_position=(0, 0),
        drone_amount=1,
        drone_speed=10,
        probability_of_detection=1,
        pre_render_time=0,
    ):
        if person_amount <= 0:
            raise ValueError("The number of persons must be greater than 0.")
        self.person_amount = person_amount

        super().__init__(
            grid_size=grid_size,
            render_mode=render_mode,
            render_grid=render_grid,
            render_gradient=render_gradient,
            timestep_limit=timestep_limit,
            drone_amount=drone_amount,
            drone_speed=drone_speed,
            probability_of_detection=probability_of_detection,
        )

        self.pre_render_steps = round(
            (pre_render_time * 60)
            / (self.calculate_simulation_time_step(drone_speed, self.cell_size))
        )
        print(f"Pre render time: {pre_render_time} minutes")
        print(f"Pre render steps: {self.pre_render_steps}")

        # Prob matrix
        self.probability_matrix = None
        self.dispersion_inc = dispersion_inc
        self.dispersion_start = dispersion_start
        self.vector = vector
        self.disaster_position = person_initial_position

        # Person initialization
        self.person_initial_position = person_initial_position
        self.persons_set = self.create_persons_set()

        # Initializing render
        self.rewards_sum = {a: 0 for a in self.possible_agents}
        self.rewards_sum["total"] = 0

        # Define the action and observation spaces for compatibility with RL libraries
        self.action_spaces = {
            agent: self.action_space(agent) for agent in self.possible_agents
        }
        self.observation_spaces = {
            agent: self.observation_space(agent) for agent in self.possible_agents
        }

    def create_persons_set(self) -> set[Person]:
        persons_set = set()
        position = self.create_random_positions_person(
            self.person_initial_position, self.person_amount
        )

        for i in range(self.person_amount):
            person = Person(
                index=i,
                initial_position=position[i],
                grid_size=self.grid_size,
            )
            person.calculate_movement_vector(self.vector)
            persons_set.add(person)

            if not self.is_valid_position(person.initial_position):
                raise ValueError("Person initial position is out of the matrix")

        return persons_set

    def create_random_positions_person(
        self, central_position: tuple[int, int], amount: int, max_distance: int = 2
    ) -> list[tuple[int, int]]:
        if not self.is_valid_position(central_position):
            raise ValueError("Central position is out of the matrix")

        max_distance_range = (max_distance * 2 + 1) ** 2

        if amount > max_distance_range:
            raise ValueError(
                "There are more persons than grid spots. Reduce number of persons or increase grid size."
            )

        unique_random_positions = {central_position}
        while len(unique_random_positions) < amount:
            dx = randint(-max_distance, max_distance)
            dy = randint(-max_distance, max_distance)

            # Checking to avoid including the central position or duplicate positions.
            if (dx, dy) != (0, 0):
                new_position = (central_position[0] + dx, central_position[1] + dy)
                if self.is_valid_position(new_position):
                    unique_random_positions.add(new_position)

        return list(unique_random_positions)

    def render(self):
        self.pygame_renderer.render_map()
        self.pygame_renderer.render_entities(self.persons_set)
        self.pygame_renderer.render_entities(self.agents_positions)
        self.pygame_renderer.refresh_screen()

    def reset(
        self,
        seed=None,
        options=None,
    ):
        vector = options.get("vector") if options else None
        self.vector = vector if vector else self.vector

        self.persons_set = self.create_persons_set()
        for person in self.persons_set:
            person.reset_position()
            person.update_time_step_relation(self.time_step_relation, self.cell_size)

        pod_multiplier = options.get("person_pod_multipliers") if options else None

        if pod_multiplier is not None:
            self.raise_if_unvalid_mult(pod_multiplier)
            for person, mult in zip(self.persons_set, pod_multiplier):
                person.set_mult(mult)
            

        self.probability_matrix = ProbabilityMatrix(
            40,
            self.dispersion_start,
            self.dispersion_inc,
            self.vector,
            [
                self.disaster_position[1],
                self.disaster_position[0],
            ],
            self.grid_size,
        )

        self.probability_matrix.update_time_step_relation(
            self.time_step_relation, self.cell_size
        )

        observations, infos = super().reset(seed=seed, options=options)
        self.rewards_sum = {a: 0 for a in self.agents}
        self.rewards_sum["total"] = 0
        return observations, infos

    def raise_if_unvalid_mult(self, individual_multiplication: list[int]) -> bool:
        if len(individual_multiplication) != len(self.persons_set):
            raise ValueError(
                "The number of multipliers must be equal to the number of persons."
            )

        for mult in individual_multiplication:
            if not isinstance(mult, (int, float)) or mult < 0:
                raise ValueError("The multiplier must be a positive number.")

    def pre_search_simulate(self):
        for _ in range(self.pre_render_steps):
            self.create_observations()
            if self.render_mode == "human":
                self.render()

    def create_observations(self):
        observations = {}

        for person in self.persons_set:
            movement_map = self.build_movement_matrix(person)
            person.step(movement_map)

        self.probability_matrix.step()

        probability_matrix = self.probability_matrix.get_matrix()
        for idx, agent in enumerate(self.agents):
            observation = (
                self.agents_positions[idx],
                probability_matrix,
            )
            observations[agent] = observation

        return observations

    def step(self, actions):
        """
        Returns a tuple with (observations, rewards, terminations, truncations, infos)
        """
        if not self._was_reset:
            raise ValueError("Please reset the env before interacting with it")

        terminations = {a: False for a in self.agents}
        rewards = {a: self.reward_scheme.default for a in self.agents}
        truncations = {a: False for a in self.agents}
        person_found = False

        for idx, agent in enumerate(self.agents):
            if agent not in actions:
                raise ValueError("Missing action for " + agent)

            drone_action = actions[agent]
            if drone_action not in self.action_space(agent):
                raise ValueError("Invalid action for " + agent)

            # Check truncation conditions (overwrites termination conditions)
            if self.timestep >= self.timestep_limit:
                rewards[agent] = self.reward_scheme.exceed_timestep
                if self.rewards_sum[agent] > 0:
                    rewards[agent] += self.rewards_sum[agent] // 2
                truncations[agent] = True
                terminations[agent] = True
                continue

            drone_x, drone_y = self.agents_positions[idx]
            is_searching = drone_action == Actions.SEARCH.value

            if drone_action != Actions.SEARCH.value:
                new_position = self.move_drone((drone_x, drone_y), drone_action)
                if not self.is_valid_position(new_position):
                    rewards[agent] = self.reward_scheme.leave_grid
                else:
                    self.agents_positions[idx] = new_position
                    rewards[agent] = self.reward_scheme.default
                self.rewards_sum[agent] += rewards[agent]
                continue

            drone_found_person = False
            for human in self.persons_set:
                drone_found_person = (
                    human.x == drone_x and human.y == drone_y and is_searching
                )
                if drone_found_person:
                    break

            random_value = random()
            if drone_found_person:
                max_detection_probability = min(human.get_mult() * self.drone.pod, 1)

                if random_value <= max_detection_probability:
                    self.persons_set.remove(human)

                    time_decay_factor = 1 - (self.timestep / self.timestep_limit)
                    time_reward_corrected = (
                        self.reward_scheme.search_and_find * time_decay_factor
                    )
                    rewards[agent] = (
                        self.reward_scheme.search_and_find + time_reward_corrected
                    )

                if len(self.persons_set) == 0:
                    person_found = True
                    for agent in self.agents:
                        terminations[agent] = True
                        truncations[agent] = True
            elif is_searching:
                prob_matrix = self.probability_matrix.get_matrix()
                rewards[agent] = (
                    prob_matrix[drone_y][drone_x]
                )

            self.rewards_sum[agent] += rewards[agent]

        self.timestep += 1
        # Get dummy infos
        infos = {drone: {"Found": person_found} for drone in self.agents}

        # CHECK COLISION - Drone
        self.compute_drone_collision(terminations, rewards)

        self.render_step(any(terminations.values()), person_found)

        # Get observations
        observations = self.create_observations()
        # If terminated, reset the agents (pettingzoo parallel env requirement)
        if any(terminations.values()) or any(truncations.values()):
            self.agents = []
        return observations, rewards, terminations, truncations, infos

    def render_step(self, terminal, person_found):
        if self.render_mode == "human":
            if terminal:
                if person_found:
                    self.pygame_renderer.render_episode_end_screen(
                        f"The target was found in {self.timestep} moves", GREEN
                    )
                else:
                    self.pygame_renderer.render_episode_end_screen(
                        "The target was not found.", RED
                    )
            else:
                self.render()

    def build_movement_matrix(self, person: Person) -> np.ndarray:
        """
        Builds and outputs a 3x3 matrix from the probabality matrix to use in the person movement function.
        """
        # Boundaries for the 3x3 movement matrix.
        left_x = max(person.x - 1, 0)
        right_x = min(person.x + 2, self.grid_size)
        left_y = max(person.y - 1, 0)
        right_y = min(person.y + 2, self.grid_size)

        probability_matrix = self.probability_matrix.get_matrix()
        movement_map = probability_matrix[left_y:right_y, left_x:right_x].copy()

        # Pad the matrix
        if person.x == 0:
            movement_map = np.insert(movement_map, 0, 0, axis=1)
        elif person.x == self.grid_size - 1:
            movement_map = np.insert(movement_map, 2, 0, axis=1)

        if person.y == 0:
            movement_map = np.insert(movement_map, 0, 0, axis=0)
        elif person.y == self.grid_size - 1:
            movement_map = np.insert(movement_map, 2, 0, axis=0)

        return movement_map

    def get_persons(self):
        return self.persons_set

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        # Observation space for each agent:
        # - MultiDiscrete: (x, y) position of the agent
        # - Box: Probability matrix
        return Tuple(
            (
                MultiDiscrete([self.grid_size, self.grid_size]),
                Box(
                    low=0,
                    high=1,
                    shape=(self.grid_size, self.grid_size),
                    dtype=np.float32,
                ),
            )
        )

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        return Discrete(len(self.possible_actions))



from DSSE.environment.simulation.dynamic_probability import ProbabilityMatrix


def create_prob_matrix(
    person_position: tuple = (10, 10),
    vector: list = [0.1, 0.1],
    grid_size: int = 40,
    disperse_inc: int = 0.1,
):
    prob_matrix = ProbabilityMatrix(
        40,
        disperse_inc,
        disperse_inc,
        vector,
        person_position,
        grid_size,
    )
    return prob_matrix


def test_matrix_dimensions():
    grid_size = 40
    prob_matrix = create_prob_matrix(grid_size=grid_size)
    assert prob_matrix.map.shape == (grid_size, grid_size)



from numpy.linalg import norm


def calculate_simulation_time_step(
    drone_max_speed: float, cell_size: float, wind_resistance: float = 0.0
) -> float:
    """
    Calculate the time step for the simulation based on the maximum speed of
    the drones and the cell size.

    Arguments:
    ----------
    drone_max_speed: float
        Maximum speed of the drones in m/s
    cell_size: float
        Size of the cells in meters
    wind_resistance: float
        Wind resistance in m/s

    Returns:
    --------
    float: The time in seconds that elapsed in real life for each simulation time step.
    """
    return cell_size / (drone_max_speed - wind_resistance)


def calculate_time_step(
    time_step: float, person_speed: tuple[float], cell_size: float
) -> float:
    """
    Parameters:
    ----------
    time_step: float
        Time step in seconds
    person_speed: tuple[float]
        Speed of the person-in-water in m/s (x and y components)
    cell_size: float
        Size of the cells in meters

    Returns:
    -------
    int
        Time step realtion in number of iterations
    """
    speed_magnitude = norm(person_speed)
    return cell_size / speed_magnitude / time_step



import numpy as np
from pettingzoo.utils.wrappers import BaseParallelWrapper
from gymnasium.spaces import Tuple, Box
from DSSE import DroneSwarmSearch


class AllPositionsWrapper(BaseParallelWrapper):
    """
    Wrapper that modifies the observation space to include the positions of all agents in all observations
    """
    def __init__(self, env: DroneSwarmSearch):
        super().__init__(env)
    
    def step(self, actions):
        obs, reward, terminated, truncated, infos = self.env.step(actions)
        self.add_other_positions_obs(obs)
        return obs, reward, terminated, truncated, infos
    
    def add_other_positions_obs(self, obs):
        prob_matrix = obs["drone0"][1]
        for idx, agent in enumerate(self.env.agents):
            agents_positions = np.array(self.env.agents_positions, dtype=np.int64)
            agents_positions[[0, idx]] = agents_positions[[idx, 0]]            
            obs[agent] = (
                agents_positions,
                prob_matrix
            )

    
    def reset(self, **kwargs):
        obs, infos = self.env.reset(**kwargs)
        self.add_other_positions_obs(obs)
        return obs, infos

    def observation_space(self, agent):
        return Tuple(
            (
                Box(0, self.env.grid_size, shape=(len(self.env.possible_agents), 2), dtype=np.int64),
                Box(
                    low=0,
                    high=1,
                    shape=(self.env.grid_size, self.env.grid_size),
                    dtype=np.float32,
                )
            )
        )
            






from DSSE import Actions
import math


def calculate_action(drone_position, target_position):
    drone_x, drone_y = drone_position
    target_x, target_y = target_position

    if drone_x == target_x and drone_y == target_y:
        return Actions.SEARCH.value

    diff_x = target_x - drone_x
    diff_y = target_y - drone_y

    if diff_x < 0 and diff_y < 0:
        return Actions.UP_LEFT.value
    elif diff_x > 0 and diff_y < 0:
        return Actions.UP_RIGHT.value
    elif diff_x < 0 and diff_y > 0:
        return Actions.DOWN_LEFT.value
    elif diff_x > 0 and diff_y > 0:
        return Actions.DOWN_RIGHT.value
    elif diff_x < 0:
        return Actions.LEFT.value
    elif diff_x > 0:
        return Actions.RIGHT.value
    elif diff_y < 0:
        return Actions.UP.value
    else:
        return Actions.DOWN.value


def distance(point1, point2):
    x1, y1 = point1
    x2, y2 = point2

    diff_x = x2 - x1
    diff_y = y2 - y1

    distance = math.sqrt(diff_x**2 + diff_y**2)
    return distance


def policy(obs, agents, env):
    actions = {}
    for idx, agent in enumerate(agents):
        castaways_positions = [person.get_position() for person in env.get_persons()]

        drone_position = env.agents_positions[idx]

        nearest_castaway = min(
            castaways_positions, key=lambda pos: distance(drone_position, pos)
        )

        action = calculate_action(drone_position, nearest_castaway)

        actions[agent] = action

    return actions



from gymnasium.spaces import Discrete
from .env_base import DroneSwarmSearchBase
from .simulation.particle_simulation import ParticleSimulation
from .constants import Reward
import numpy as np
import functools


class CoverageDroneSwarmSearch(DroneSwarmSearchBase):
    metadata = {
        "name": "DroneSwarmSearchCPP",
    }
    reward_scheme = Reward(
        default=0,
        leave_grid=-10,
        exceed_timestep=-100,
        drones_collision=-10,
        search_cell=10,
        search_and_find=100,
    )

    def __init__(
        self,
        render_mode="ansi",
        render_grid=True,
        render_gradient=True,
        timestep_limit=100,
        disaster_position=(-24.04, -46.17),
        drone_amount=1,
        drone_speed=10,
        drone_probability_of_detection=0.9,
        pre_render_time=10,
        prob_matrix_path=None,
    ) -> None:

        # Prob matrix
        self.probability_matrix = ParticleSimulation(
            disaster_lat=disaster_position[0],
            disaster_long=disaster_position[1],
            duration_hours=pre_render_time,
        )
        if prob_matrix_path is not None:
            if not isinstance(prob_matrix_path, str):
                raise ValueError("prob_matrix_path must be a string")
            self.probability_matrix.load_state(prob_matrix_path)
        else: 
            self.probability_matrix.run_or_get_simulation()
        grid_size = self.probability_matrix.get_map_size()

        super().__init__(
            grid_size=grid_size,
            render_mode=render_mode,
            render_grid=render_grid,
            render_gradient=render_gradient,
            timestep_limit=timestep_limit,
            drone_amount=drone_amount,
            drone_speed=drone_speed,
            probability_of_detection=drone_probability_of_detection,
        )
        self.disaster_position = disaster_position
        # Sets used to keep track of the seen and not seen states for reward calculation
        self.seen_states = None
        self.not_seen_states = None
        self.all_states = {
            (x, y) for x in range(self.grid_size) for y in range(self.grid_size)
        }
        self.repeated_coverage = 0
        self.cumm_pos = 0

    def reset(self, seed=None, options=None):
        obs, _ = super().reset(seed=seed, options=options)

        self.reset_search_state()

        self.cumm_pos = 0
        self.repeated_coverage = 0
        infos = self.compute_infos(False)
        return obs, infos

    def reset_search_state(self):
        # This is in (x, y)
        self.seen_states = {pos for pos in self.agents_positions}

        mat = self.probability_matrix.get_matrix()
        # (row, col)
        close_to_zero = np.argwhere(np.abs(mat) < 1e-10)

        # Remove the need to visit cells with POC near to 0
        for y, x in close_to_zero:
            self.seen_states.add((x, y))

        self.not_seen_states = self.all_states - self.seen_states

    def create_observations(self):
        observations = {}

        probability_matrix = self.probability_matrix.get_matrix()
        for idx, agent in enumerate(self.agents):
            observation = (
                self.agents_positions[idx],
                probability_matrix,
            )
            observations[agent] = observation

        return observations

    def pre_search_simulate(self):
        self.probability_matrix.run_or_get_simulation()

    def step(self, actions: dict[str, int]) -> tuple:
        if not self._was_reset:
            raise ValueError("Please reset the env before interacting with it")

        terminations = {a: False for a in self.agents}
        rewards = {a: self.reward_scheme.default for a in self.agents}
        truncations = {a: False for a in self.agents}
        self.timestep += 1

        prob_matrix = self.probability_matrix.get_matrix()
        for idx, agent in enumerate(self.agents):
            if agent not in actions:
                raise ValueError("Missing action for " + agent)

            drone_action = actions[agent]
            if drone_action not in self.action_space(agent):
                raise ValueError("Invalid action for " + agent)


            if self.timestep >= self.timestep_limit:
                rewards[agent] = self.reward_scheme.exceed_timestep
                truncations[agent] = True
                continue

            # Action 8 is to stay in the same position, default reward.
            if drone_action == 8:
                continue
        
            drone_x, drone_y = self.agents_positions[idx]
            new_position = self.move_drone((drone_x, drone_y), drone_action)
            if not self.is_valid_position(new_position):
                rewards[agent] = self.reward_scheme.leave_grid
                continue

            self.agents_positions[idx] = new_position
            new_x, new_y = new_position
            if new_position in self.not_seen_states:
                reward_poc = (1 / (self.timestep)) * prob_matrix[new_y, new_x] * 1_000
                rewards[agent] = self.reward_scheme.search_cell + reward_poc
                self.seen_states.add(new_position)
                self.not_seen_states.remove(new_position)
                # Probability of sucess (POS) = POC * POD
                self.cumm_pos += prob_matrix[new_y, new_x] * self.drone.pod
                # Remove the probability of the visited cell.
                prob_matrix[new_y, new_x] = 0.0
            else:
                self.repeated_coverage += 1

        # Get dummy infos
        is_completed = len(self.not_seen_states) == 0
        if self.render_mode == "human":
            self.render()

        if is_completed:
            # TODO: Proper define reward for completing the search (R_done)
            rewards = {
                drone: self.reward_scheme.search_and_find for drone in self.agents
            }
            terminations = {drone: True for drone in self.agents}
        infos = self.compute_infos(is_completed)

        self.compute_drone_collision(terminations, rewards)
        # Get observations
        observations = self.create_observations()
        # If terminted, reset the agents (pettingzoo parallel env requirement)
        if any(terminations.values()) or any(truncations.values()):
            self.agents = []
        return observations, rewards, terminations, truncations, infos

    def compute_infos(self, is_completed: bool) -> dict[str, dict]:
        # TODO: Is this the best way to inform the coverage rate, Cum_pos and repetitions?
        total_states = len(self.seen_states) + len(self.not_seen_states)
        coverage_rate = len(self.seen_states) / total_states
        infos = {
            "is_completed": is_completed,
            "coverage_rate": coverage_rate,
            "repeated_coverage": self.repeated_coverage / total_states,
            "acumulated_pos": self.cumm_pos,
        }
        return {drone: infos for drone in self.agents}
    
    def save_matrix(self, path: str):
        self.probability_matrix.save_state(path)

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        return Discrete(9)



import numpy as np
from random import uniform, choice
from numpy.linalg import norm
from math import cos, sin, radians, pi, exp
from DSSE.environment.simulation.time_step import calculate_time_step


class Person:
    """
    Class that represents a shipwrecked person in the environment.
    """

    def __init__(
        self,
        index: int,
        initial_position: tuple[int, int],
        grid_size: int,
        mult: float = 1,
    ):
        """
        Class that represents a shipwrecked person in the environment.

        Parameters:
        -----------
        initial_position: tuple
            The initial position of the shipwrecked person in the environment.
        grid_size: int
            The size of the grid that represents the environment (grid_size x grid_size).
        mult: float, optional
            The multiplication factor that adjusts the probability of the person's finding.
        """
        self.index = index
        self.initial_position = initial_position
        self.x, self.y = self.initial_position
        self.inc_x, self.inc_y = 0, 0
        self.grid_size = grid_size
        self.time_step_relation = 1
        self.mult = mult
        self.movement_vector = (0.0, 0.0)
        self.angle_range = choice([(0, 45), (20, 55), (300, 360), (315, 360)])

    def step(self, movement_map: np.array) -> None:
        if self.will_move():
            movement = self.update_position(
                movement_map, dimension=movement_map.shape[0], prob_walk_weight=0.05
            )
            self.safe_position_update(movement)

    def update_position(
        self,
        movement_map: np.array,
        dimension: int = 3,
        prob_walk_weight: float = 1.0,
    ) -> tuple[int]:
        """
        Function that takes a cut of the DynamicProbability matrix,
        along with the dimension of the matrix and the probability of walking,
        that will be used to wheight the decision of walking towards the person's movement vector.

        Output:
            (movement_x, movement_y): tuple[int]
        """
        self.increment_movement()

        # On row, column notation (y, x) -> (row, column)
        movement = (
            round(self.inc_x),
            round(self.inc_y),
        )

        row_mov = np.sign(movement[1]) + 1
        col_mov = np.sign(movement[0]) + 1
        movement_map[row_mov][col_mov] += prob_walk_weight
        movement_map /= np.sum(movement_map)
        movement_index = np.random.choice(9, size=1, p=movement_map.flatten())[0]

        if abs(self.inc_x) >= 1:
            self.inc_x -= np.sign(self.inc_x)
        if abs(self.inc_y) >= 1:
            self.inc_y -= np.sign(self.inc_y)

        movement_cartesian = self.movement_to_cartesian(
            movement_index, dimension=dimension
        )
        return movement_cartesian

    def movement_to_cartesian(
        self, movement_index: int, dimension: int = 3
    ) -> tuple[int]:
        """
        The movement of the shipwrecked person on the input follows the scheme
        (for the value of line and column):
            - if 0 -> Move to the left (x - 1) or to the top (y - 1).
            - if 1 -> No movement.
            - if 2 -> Move to the right (x + 1) or to the bottom (y + 1).

        So this function converts from this matrix movement notation to cartesian,
        as the matrix that creates this indexes is only 3x3,
        just removing 1 converts it back to cartesian movement.
        """
        x_component = (movement_index % dimension) - (dimension // 2)
        y_component = (movement_index // dimension) - (dimension // 2)

        return x_component, y_component

    def will_move(self) -> bool:
        if abs(self.inc_x) >= 1 or abs(self.inc_y) >= 1:
            return True
        self.increment_movement()
        return False

    def increment_movement(self) -> None:
        self.inc_x += self.movement_vector[0] / self.time_step_relation
        self.inc_y += self.movement_vector[1] / self.time_step_relation

    def calculate_movement_vector(self, primary_movement_vector: tuple[float]) -> None:
        """
        Function that calculates the person's movement vector
        based on the primary movement vector that is being applied
        by the environment, that is the water drift vector.

        The resulting movement vector is the average of the primary movement vector
        and a noised version of it, simulating more natural or unpredictable motion.
        """
        noised_vector = self.noise_vector(primary_movement_vector)
        self.movement_vector = (
            (primary_movement_vector[0] + noised_vector[0]) / 2,
            (primary_movement_vector[1] + noised_vector[1]) / 2,
        )

    def noise_vector(
        self,
        primary_movement_vector: tuple[float],
        speed_factor_range: tuple[float, float] = (0.5, 1.5),
    ) -> tuple[float]:
        """
        Generates a 'noised' version of a given primary movement vector,
        simulating more natural or unpredictable motion.

        The function introduces variability to the vector's direction and magnitude by
        applying a randomly chosen angle and an inverse exponential adjustment based on
        the angle's deviation from a base angle (90 degrees).
        A speed factor, influenced by the angle's normalized deviation, scales the
        vector's magnitude within a specified range.

        Parameters:
        ----------
        primary_movement_vector : tuple[float]
            The original movement vector to be noised.
        speed_factor_range : tuple[float, float], optional
            A tuple specifying the range within which the speed factor will be scaled.
            Default values allows for both reduction and amplification of the vector's magnitude.

        Returns:
        -------
        tuple[float]: A noised version of the input vector, represented as a tuple of floats,
        incorporating random variations in both direction and magnitude.

        Notes:
        ------
        If the input vector is a zero vector, the function returns a zero vector since no direction
        or magnitude can be meaningfully adjusted.
        """

        if norm(primary_movement_vector) == 0:
            return (0.0, 0.0)

        angle = radians(uniform(*self.angle_range))
        angle_base = pi / 2
        normalized_angle_diff = abs(angle - angle_base) / pi

        speed_factor = speed_factor_range[0] + (
            exp(-normalized_angle_diff)
            * (speed_factor_range[1] - speed_factor_range[0])
        )

        primary_direction = np.array(primary_movement_vector) / norm(
            primary_movement_vector
        )
        rotation_matrix = np.array(
            [[cos(angle), -sin(angle)], [sin(angle), cos(angle)]]
        )
        noised_direction = np.dot(rotation_matrix, primary_direction)
        noised_vector = noised_direction * norm(primary_movement_vector) * speed_factor

        return tuple(noised_vector)

    def safe_position_update(self, movement: tuple[int]) -> None:
        """
        Updates the shipwrecked person position on given movement, checking for edge cases first.
        """
        new_x = self.x + movement[0]
        new_y = self.y + movement[1]
        if self.is_safe_position(new_x):
            self.x = new_x
        if self.is_safe_position(new_y):
            self.y = new_y

    def is_safe_position(self, new_position: int) -> bool:
        return 0 <= new_position < self.grid_size

    def update_time_step_relation(self, time_step: float, cell_size: float) -> None:
        self.time_step_relation = calculate_time_step(
            time_step, self.movement_vector, cell_size
        )

    def reset_position(self):
        self.x, self.y = self.initial_position

    def get_position(self) -> tuple[int]:
        return (self.x, self.y)

    def set_mult(self, mult: int) -> None:
        self.mult = mult

    def get_mult(self) -> int:
        return self.mult

    def __hash__(self) -> int:
        return hash(self.index)



import time
import pygame
import os
import numpy as np
from .constants import BLACK


class PygameInterface:
    """
    Class for rendering the grafical interface of the simulation
    """

    FPS = 5

    def __init__(
        self, grid_size: int, render_gradient: bool, render_grid: bool
    ) -> None:
        pygame.init()
        self.grid_size = grid_size
        self.render_gradient = render_gradient
        self.render_grid = render_grid
        self.window_size = 700
        self.screen = None
        self.render_on = False
        self.probability_matrix = None

        self.block_size = self.window_size / self.grid_size
        self.drone_img = None
        self.person_img = None
        self.clock = None

    def render_map(self):
        self.draw()

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.close()

        pygame.event.pump()

    def refresh_screen(self):
        pygame.display.flip()

    def enable_render(self):
        if self.render_on:
            return

        self.screen = pygame.Surface([self.window_size + 20, self.window_size + 20])
        self.screen = pygame.display.set_mode(self.screen.get_size())

        current_directory = os.path.dirname(os.path.abspath(__file__))
        self.drone_img = self.load_and_scale_image(
            f"{current_directory}/imgs/drone.png"
        )
        self.person_img = self.load_and_scale_image(
            f"{current_directory}/imgs/person-swimming.png"
        )

        self.clock = pygame.time.Clock()
        self.render_on = True

    def load_and_scale_image(self, image_path: str):
        loaded_img = pygame.image.load(image_path).convert()
        scaled_img = pygame.transform.scale(
            loaded_img, (self.block_size, self.block_size)
        )
        return scaled_img

    def render_entities(self, entities) -> None:
        for entity in entities:
            # Checks if the entity is a tuple (assuming that drones are represented as tuples).
            if isinstance(entity, tuple):
                rectangle = self.get_position_rectangle(entity)
                image = self.drone_img
            # Otherwise, assumes it is an object with 'x' and 'y' attributes (such as a person).
            else:
                rectangle = self.get_position_rectangle((entity.x, entity.y))
                image = self.person_img

            # Renders the entity.
            self.screen.blit(image, rectangle)

    def get_position_rectangle(self, position: tuple[int, int]) -> pygame.Rect:
        x = 10 + self.block_size * position[0]
        y = 10 + self.block_size * position[1]
        return pygame.Rect(x, y, self.block_size, self.block_size)

    def draw(self):
        self.clock.tick(self.FPS)
        self.screen.fill(BLACK)

        matrix = self.probability_matrix.get_matrix()
        max_matrix = matrix.max()
        if max_matrix == 0.0:
            max_matrix = 1.0

        for counter_x, x in enumerate(
            np.arange(10, self.window_size + 10, self.block_size)
        ):
            for counter_y, y in enumerate(
                np.arange(10, self.window_size + 10, self.block_size)
            ):
                rectangle = pygame.Rect(x, y, self.block_size, self.block_size)
                prob = matrix[counter_y][counter_x]
                normalized_prob = prob / max_matrix

                computed_prob_color = self.compute_cell_color(normalized_prob)
                pygame.draw.rect(self.screen, computed_prob_color, rectangle)

                if self.render_grid:
                    pygame.draw.rect(self.screen, BLACK, rectangle, 2)

    def compute_cell_color(self, normalized_prob):
        if self.render_gradient:
            green = normalized_prob * 255
            red = (1 - normalized_prob) * 255
            max_color = max(red, green)
            green = (green * 255) / (max_color)
            red = (red * 255) / (max_color)
        else:
            red = 255
            green = 0
            if normalized_prob >= 0.75:
                red = 0
                green = 255
            elif normalized_prob >= 0.25:
                red = 255
                green = 255
        return (red, green, 0)

    def render_episode_end_screen(self, message: str, color: tuple):
        font = pygame.font.SysFont(None, 50)
        text = font.render(message, True, BLACK)
        text_rect = text.get_rect(center=(self.window_size // 2, self.window_size // 2))
        self.screen.fill(color)
        self.screen.blit(text, text_rect)
        pygame.display.flip()
        time.sleep(1)

    def close(self):
        if self.render_on:
            pygame.event.pump()
            pygame.display.quit()
            self.render_on = False



import functools
import numpy as np
from abc import ABC, abstractmethod
from pettingzoo import ParallelEnv
from .entities.drone import DroneData
from .pygame_interface import PygameInterface
from .simulation.dynamic_probability import ProbabilityMatrix
from .constants import Actions
from gymnasium.spaces import MultiDiscrete, Discrete, Tuple, Box
from copy import copy


class DroneSwarmSearchBase(ABC, ParallelEnv):
    def __init__(
        self,
        grid_size=7,
        render_mode="ansi",
        render_grid=False,
        render_gradient=True,
        timestep_limit=100,
        drone_amount=1,
        drone_speed=10,
        probability_of_detection=1,
    ) -> None:
        self.cell_size = 130  # in meters
        self.grid_size = grid_size
        self._was_reset = False
        if not isinstance(drone_amount, int):
            raise ValueError("Drone amount must be an integer")
    
        self.drone = DroneData(
            amount=drone_amount,
            speed=drone_speed,
            pod=probability_of_detection,
        )
        self.probability_of_detection = probability_of_detection

        # Error Checking
        if self.drone.amount > self.grid_size * self.grid_size:
            raise ValueError(
                "There are more drones than grid spots. Reduce number of drones or increase grid size."
            )

        if render_mode != "ansi" and render_mode != "human":
            raise ValueError("Render mode not recognized")

        self.timestep = None
        self.timestep_limit = timestep_limit
        self.time_step_relation = self.calculate_simulation_time_step(
            self.drone.speed, self.cell_size
        )

        self.possible_agents = []
        self.agents_positions = [(None, None)] * self.drone.amount
        for i in range(self.drone.amount):
            agent_name = "drone" + str(i)
            self.possible_agents.append(agent_name)

        self.render_mode = render_mode

        # Initializing render
        self.pygame_renderer = PygameInterface(
            self.grid_size, render_gradient, render_grid
        )

    def calculate_simulation_time_step(
        self, drone_max_speed: float, cell_size: float, wind_resistance: float = 0.0
    ) -> float:
        """
        Calculate the time step for the simulation based on the maximum speed of the drones and the cell size

        Args:
        max_speed: float
            Maximum speed of the drones in m/s
        cell_size: float
            Size of the cells in meters
        wind_resistance: float
            Wind resistance in m/s
        """
        return cell_size / (drone_max_speed - wind_resistance)  # in seconds

    def render(self):
        self.pygame_renderer.render_map()
        self.pygame_renderer.render_entities(self.agents_positions)
        self.pygame_renderer.refresh_screen()

    @abstractmethod
    def reset(
        self,
        seed=None,
        options=None,
    ):
        self._was_reset = True
        drones_positions = options.get("drones_positions") if options else None

        if drones_positions is not None:
            if not self.is_valid_position_drones(drones_positions):
                raise ValueError(
                    "You are trying to place the drone in a invalid position"
                )

        self.agents = copy(self.possible_agents)
        self.timestep = 0

        if drones_positions is None:
            self.default_drones_positions()
        else:
            self.required_drone_positions(drones_positions)

        if self.render_mode == "human":
            self.pygame_renderer.probability_matrix = self.probability_matrix
            self.pygame_renderer.enable_render()
            self.render()

        self.pre_search_simulate()
        observations = self.create_observations()
        infos = {drone: {"Found": False} for drone in self.agents}
        return observations, infos

    def is_valid_position_drones(self, positions: list[tuple[int, int]]) -> bool:
        seen = set()
        for position in positions:
            if not self.is_valid_position(position) or position in seen:
                return False
            seen.add(position)
        return True

    def is_valid_position(self, position: tuple[int, int]) -> bool:
        valid_x = position[0] >= 0 and position[0] < self.grid_size
        valid_y = position[1] >= 0 and position[1] < self.grid_size
        return valid_x and valid_y

    def default_drones_positions(self):
        counter_x = 0
        counter_y = 0
        for agent_index in range(len(self.agents)):
            if counter_x >= self.grid_size:
                counter_x = 0
                counter_y += 1
            self.agents_positions[agent_index] = (counter_x, counter_y)
            counter_x += 1

    def required_drone_positions(self, drones_positions: list):
        if len(drones_positions) != len(self.possible_agents):
            raise ValueError(
                "There are more or less initial positions than drones,"
                "please make sure there are the same number of initial possitions "
                "and number of drones."
            )
        for i in range(len(drones_positions)):
            x, y = drones_positions[i]
            self.agents_positions[i] = (x, y)

    @abstractmethod
    def pre_search_simulate(self):
        raise NotImplementedError("Method not implemented")

    @abstractmethod
    def create_observations(self):
        raise NotImplementedError("Method not implemented")

    @abstractmethod
    def step(self, actions):
        raise NotImplementedError("Method not implemented")

    def compute_drone_collision(self, terminations, rewards):
        """
        Check for drone collision and compute terminations, rewards and truncations.
        """
        for drone_1_id in range(len(self.agents)):
            for drone_2_id in range(drone_1_id + 1, len(self.agents)):
                drone_1_name = self.agents[drone_1_id]
                drone_2_name = self.agents[drone_2_id]
                if self.agents_positions[drone_1_id] == self.agents_positions[drone_2_id]:
                    terminations[drone_1_name] = True
                    terminations[drone_2_name] = True
                    rewards[drone_1_name] = self.reward_scheme.drones_collision
                    rewards[drone_2_name] = self.reward_scheme.drones_collision


    def move_drone(self, position, action):
        """
        Returns a tuple with (is_terminal, new_position, reward)
        """
        match action:
            case Actions.LEFT.value:  # LEFT
                new_position = (position[0] - 1, position[1])
            case Actions.RIGHT.value:  # RIGHT
                new_position = (position[0] + 1, position[1])
            case Actions.UP.value:  # UP
                new_position = (position[0], position[1] - 1)
            case Actions.DOWN.value:  # DOWN
                new_position = (position[0], position[1] + 1)
            case Actions.UP_LEFT.value:  # UP_LEFT
                new_position = (position[0] - 1, position[1] - 1)
            case Actions.UP_RIGHT.value:  # UP_RIGHT
                new_position = (position[0] + 1, position[1] - 1)
            case Actions.DOWN_LEFT.value:  # DOWN_LEFT
                new_position = (position[0] - 1, position[1] + 1)
            case Actions.DOWN_RIGHT.value:  # DOWN_RIGHT
                new_position = (position[0] + 1, position[1] + 1)

        return new_position

    def get_agents(self):
        return self.possible_agents

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        # Observation space for each agent:
        # - MultiDiscrete: (x, y) position of the agent
        # - Box: Probability matrix
        return Tuple(
            (
                MultiDiscrete([self.grid_size, self.grid_size]),
                Box(
                    low=0,
                    high=1,
                    shape=(self.grid_size, self.grid_size),
                    dtype=np.float32,
                ),
            )
        )

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        return Discrete(len(self.possible_actions))



import pytest
from DSSE import CoverageDroneSwarmSearch
from DSSE.environment.constants import Actions
from pettingzoo.test import parallel_api_test
import numpy as np


def init_Coverage_drone_swarm_search(
    render_mode="ansi",
    render_grid=True,
    render_gradient=True,
    timestep_limit=100,
    drone_amount=1,
    pre_render_time=10,
    prob_matrix_path="DSSE/tests/matrix.npy",
    disaster_position=(-24.04, -46.17),
    drone_speed=10,
    drone_probability_of_detection=1,
):
    
    return CoverageDroneSwarmSearch(
        render_mode=render_mode,
        render_grid=render_grid,
        render_gradient=render_gradient,
        timestep_limit=timestep_limit,
        drone_amount=drone_amount,
        pre_render_time=pre_render_time,
        prob_matrix_path=prob_matrix_path,
        disaster_position=disaster_position,
        drone_speed=drone_speed,
        drone_probability_of_detection=drone_probability_of_detection,
    )


@pytest.mark.parametrize(
    "drone_amount",
    [
        -26,
        101.2,
        "401",
        4901,
    ],
)
def test_wrong_drone_number(drone_amount):
    with pytest.raises(ValueError):
        init_Coverage_drone_swarm_search(drone_amount=drone_amount)


@pytest.mark.parametrize(
    "drone_amount, drones_positions",
    [
        (2, [(0, 0), (2, 0),]),
    ],
)
def test_drone_collision_termination(drone_amount, drones_positions):

    env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)
    opt = {
    "drones_positions": drones_positions,
    }
    _ = env.reset(options=opt)

    done = False
    while not done:
        actions = {"drone0": Actions.RIGHT.value, "drone1": Actions.LEFT.value}
        _, reward, terminations, truncations, _ = env.step(actions)
        done = any(terminations.values()) or any(truncations.values())

        assert done, "The simulation should terminate upon drone collision."
        assert any(
            terminations.values()
        ), "There should be a termination flag set due to the collision."
        assert (
            sum(reward.values()) < 0
        ), "The total reward should be negative after a collision."

@pytest.mark.parametrize(
    "drone_amount, drones_positions",
    [
        (1, [(0, 0)]),
    ],
)
def test_leave_grid_get_negative_reward(drone_amount, drones_positions):
    env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)
    opt = {"drones_positions": drones_positions}
    _ = env.reset(options=opt)

    done = False
    reward_sum = 0
    while not done and reward_sum >=  (env.reward_scheme.leave_grid * (env.timestep_limit-1)) +1:
        actions = {"drone0": Actions.UP.value}
        _, reward, terminations, done, _ = env.step(actions)
        done = any(done.values())
        reward_sum += sum(reward.values())

    assert (
        not done
    ), "The simulation should not end, indicating the drone left the grid or another termination condition was met."
    assert (
        sum(reward.values()) < 0
    ), "The total reward should be negative, indicating a penalty was applied."
    assert not any(
        terminations.values()
    ), "There not should be at least one termination condition met."

@pytest.mark.parametrize(
    "drone_amount",
    [
        5,
        10,
        15,
        20,
    ],
)
def test_if_all_drones_are_created(drone_amount):
    env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)
    _ = env.reset()

    assert (
        len(env.get_agents()) == drone_amount
    ), f"Should have {drone_amount} drones, but found {len(env.get_agents())}."


@pytest.mark.parametrize(
    "drone_amount, drones_positions",
    [
        (1, [(3, 3)]),
        (2, [(12, 0), (0, 13)]),
        (3, [(0, 0), (19, 19), (15, 10)]),
        (4, [(5, 0), (0, 0), (1, 1), (10, 10)]),
    ],
)
def test_position_drone_is_correct_after_reset(drone_amount, drones_positions):
    env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)

    opt = {"drones_positions": drones_positions}
    observations, _ = env.reset(options=opt)

    for i, position in enumerate(drones_positions):
        drone_id = f"drone{i}"
        real_position_drone = observations[drone_id][0]

        assert (
            real_position_drone == position
        ), f"Expected {drone_id}'s position to be {position} after reset, but was {real_position_drone}."


@pytest.mark.parametrize(
    "drone_amount, drones_positions",
    [
        (1, [(-1, 3)]),
        (2, [(1200, 0), (25, 13)]),
        (3, [(0, 0), (19, 19), (25, -10)]),
        (4, [(5, 0), (0, 0), (10, 10), (10, 10)]),
    ],
)
def test_invalid_drone_position_raises_error(drone_amount, drones_positions):
    with pytest.raises(ValueError):
        env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)
        opt = {"drones_positions": drones_positions}
        _ = env.reset(options=opt)


@pytest.mark.parametrize(
    "drone_amount",
    [
        1,
        20,
        35,
        48,
    ],
)
def test_if_all_drones_are_created_with_default_positions(drone_amount):
    env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)

    _ = env.reset()

    assert (
        len(env.get_agents()) == drone_amount
    ), f"Should have {drone_amount} drones, but found {len(env.get_agents())}."


@pytest.mark.parametrize(
    "drone_amount",
    [
        2,
        5,
        15,
    ],
)
def test_with_the_observation_size_is_correct_for_all_drones(drone_amount):
    env = init_Coverage_drone_swarm_search(drone_amount=drone_amount)

    observations, _ = env.reset()

    # Obtém o tamanho da primeira observação
    first_observation_shape = observations[f"drone0"][1].shape

    for drone in range(1, drone_amount):
        drone_id = f"drone{drone}"
        observation_matriz = observations[drone_id][1]
        
        # Verifica se o tamanho da observação é igual ao da primeira observação
        assert np.array_equal(observation_matriz.shape, first_observation_shape), f"Observation size mismatch for drone {drone_id}. Expected size: {first_observation_shape}, actual size: {observation_matriz.shape}"

@pytest.mark.parametrize(
    "timestep_limit",
    [
        100,
        550,
        150,
    ],
)
def test_if_the_timestep_limit_is_correct(timestep_limit):
    # Cria o ambiente com o limite de passos de tempo especificado
    env = init_Coverage_drone_swarm_search(timestep_limit=timestep_limit)
    _, _ = env.reset()

    # Contador de passos de tempo
    steps = 0

    # Loop até o ambiente atingir o limite de passos de tempo
    while env.agents:
        steps += 1
        actions = {agent: env.action_space(agent).sample() for agent in env.agents}
        _, _, _, _, _ = env.step(actions)
    
    assert steps == timestep_limit, f"Number of steps does not match timestep limit. Expected: {timestep_limit}, Actual: {steps}"
    

def test_petting_zoo_interface_works():
    env = init_Coverage_drone_swarm_search()
    parallel_api_test(env)
    env.close()



from enum import Enum
from collections import namedtuple

# Colors
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLACK = (0, 0, 0)


# Actions
class Actions(Enum):
    LEFT = 0
    RIGHT = 1
    UP = 2
    DOWN = 3
    UP_LEFT = 4
    UP_RIGHT = 5
    DOWN_LEFT = 6
    DOWN_RIGHT = 7
    SEARCH = 8


Reward = namedtuple(
    "Reward",
    [
        "default",
        "leave_grid",
        "exceed_timestep",
        "drones_collision",
        "search_cell",
        "search_and_find",
    ],
)



import math
import numpy as np
from datetime import datetime, timedelta
from typing import List, Tuple


EARTH_MEAN_RADIUS = 6373.0


class ParticleSimulation:
    def __init__(
        self,
        disaster_lat: float,
        disaster_long: float,
        duration_hours: int = 10,
        loglevel: int = 20,
        animate: bool = False,
        cell_size: int = 130,
    ) -> None:
        try:
            from opendrift.models.oceandrift import OceanDrift
            self.ocean_drift = OceanDrift
        except ImportError:
            raise ImportError("OpenDrift not installed. Install the environment with the 'coverage' extra: pip install DSSE[coverage]")
        
        self.disaster_lat = disaster_lat
        self.disaster_long = disaster_long
        self.loglevel = loglevel
        self.animate = animate
        self.duration_hours = duration_hours
        self.cell_size = cell_size

        # Internal variables
        self.map_size = 0
        self.original_map = None
        self.probability_map = None

    def run_or_get_simulation(self):
        if self.probability_map is None:
            self.run_simulation()
        self.probability_map = self.original_map.copy()

    def run_simulation(self):
        duration = timedelta(hours=self.duration_hours)
        start_time = datetime.now() - duration
        number = 50_000
        radius = 1000

        coordinates = self.simulate(start_time, number, radius, duration)
        self.map_size = self.calculate_map_size(coordinates)
        cartesian = self.convert_lat_lon_to_xy(coordinates)
        self.probability_map = self.create_probability_map(cartesian)
        # Maintain always a copy of the original map
        self.original_map = self.probability_map.copy()

    def simulate(
        self,
        time: datetime,
        number: int,
        radius: int,
        duration: timedelta,
    ) -> List[Tuple[float, float]]:
        o = self.ocean_drift(loglevel=self.loglevel)
        o.add_readers_from_list(
            ["https://tds.hycom.org/thredds/dodsC/GLBy0.08/expt_93.0/uv3z"]
        )
        o.seed_elements(
            lat=self.disaster_lat,
            lon=self.disaster_long,
            time=time,
            number=number,
            radius=radius,
        )

        o.run(duration=duration, time_step=1800)
        if self.animate:
            o.animation(filename="animation.mp4")

        lat_final = o.elements.lat
        lon_final = o.elements.lon

        return list(zip(lat_final, lon_final))

    def lat_long_to_global_xy(
        self, lat: float, lon: float, ratio: float
    ) -> Tuple[int, int]:
        lat = math.radians(lat)
        lon = math.radians(lon)

        x = EARTH_MEAN_RADIUS * lon * ratio
        y = EARTH_MEAN_RADIUS * lat
        return (x, y)

    def convert_lat_lon_to_xy(
        self, coordinates: List[Tuple[float, float]]
    ) -> List[Tuple[int, int]]:
        """
        Using equirectangular projection to convert latitudes and longitudes to Cartesian coordinates.
        """
        min_lat, max_lat, min_lon, max_lon = self.calculate_bounding_rectangle(
            coordinates
        )

        ratio = math.cos((math.radians(min_lat) + math.radians(max_lat)) / 2)
        p0_x, p0_y = self.lat_long_to_global_xy(min_lat, min_lon, ratio)
        p1_x, p1_y = self.lat_long_to_global_xy(max_lat, max_lon, ratio)

        cartesian_coordinates = []
        for lat, lon in coordinates:
            x, y = self.lat_long_to_global_xy(lat, lon, ratio)
            percent_x = (x - p0_x) / (p1_x - p0_x)
            percent_y = (y - p0_y) / (p1_y - p0_y)
            actual_x = (self.map_size - 1) * percent_x
            actual_y = (self.map_size - 1) * percent_y
            cartesian_coordinates.append((round(actual_x), round(actual_y)))
        return cartesian_coordinates

    def calculate_map_size(self, coordinates) -> int:

        min_lat, max_lat, min_lon, max_lon = self.calculate_bounding_rectangle(
            coordinates
        )

        width = self.distance_in_meters(min_lat, min_lon, min_lat, max_lon)
        height = self.distance_in_meters(min_lat, min_lon, max_lat, min_lon)

        map_width = round(width / self.cell_size)
        map_height = round(height / self.cell_size)

        return max(map_width, map_height)

    def calculate_bounding_rectangle(
        self, coordinates: List[Tuple[float, float]]
    ) -> Tuple[float, float, float, float]:
        latitudes, longitudes = zip(*coordinates)

        min_lat, max_lat = min(latitudes), max(latitudes)
        min_lon, max_lon = min(longitudes), max(longitudes)

        return min_lat, max_lat, min_lon, max_lon

    def distance_in_meters(
        self, lat1: float, lon1: float, lat2: float, lon2: float
    ) -> float:
        """
        Uses the Haversine formula to calculate the distance between two points on Earth, given their latitudes and longitudes.

        Returns the distance in meters.
        """
        lat1 = math.radians(lat1)
        lon1 = math.radians(lon1)
        lat2 = math.radians(lat2)
        lon2 = math.radians(lon2)

        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = (
            math.sin(dlat / 2) ** 2
            + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
        )
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

        return EARTH_MEAN_RADIUS * c * 1000

    def create_probability_map(
        self, cartesian_coords: List[Tuple[float, float]]
    ) -> List[List[int]]:
        """
        Creates a probability map based on the coordinates of the particles.
        """
        prob_map = np.zeros((self.map_size, self.map_size))

        for x, y in cartesian_coords:
            prob_map[y][x] += 1

        particle_sum = max(np.sum(prob_map), 1)

        probability_map = prob_map / particle_sum
        return probability_map

    def get_matrix(self):
        return self.probability_map

    def get_map_size(self):
        return self.map_size
    
    def save_state(self, output_path: str):
        with open(output_path, "wb") as f:
            np.save(f, self.original_map)
        
    
    def load_state(self, input_path: str):
        with open(input_path, "rb") as f:
            self.probability_map = np.load(f)
            self.original_map = self.probability_map.copy()
            self.map_size = len(self.probability_map)


import pytest
from DSSE import DroneSwarmSearch
from DSSE.environment.constants import Actions
from DSSE.tests.drone_policy import policy
from pettingzoo.test import parallel_api_test


def init_drone_swarm_search(
    grid_size=20,
    render_mode="ansi",
    render_grid=True,
    render_gradient=True,
    vector=(3.5, -0.5),
    disperse_inc=0.1,
    timestep_limit=200,
    person_amount=1,
    person_initial_position=None,
    drone_amount=1,
    drone_speed=10,
    probability_of_detection=0.9,
    pre_render_time=0,
):

    if person_initial_position is None:
        person_initial_position = (
            grid_size - round(grid_size / 2),
            grid_size - round(grid_size / 2),
        )

    return DroneSwarmSearch(
        grid_size=grid_size,
        render_mode=render_mode,
        render_grid=render_grid,
        render_gradient=render_gradient,
        vector=vector,
        dispersion_inc=disperse_inc,
        timestep_limit=timestep_limit,
        person_amount=person_amount,
        person_initial_position=person_initial_position,
        drone_amount=drone_amount,
        drone_speed=drone_speed,
        probability_of_detection=probability_of_detection,
        pre_render_time=pre_render_time,
    )


@pytest.mark.parametrize(
    "grid_size, drone_amount",
    [
        (5, 26),
        (10, 101),
        (20, 401),
        (50, 2501),
    ],
)
def test_wrong_drone_number(grid_size, drone_amount):
    with pytest.raises(ValueError):
        init_drone_swarm_search(grid_size=grid_size, drone_amount=drone_amount)


@pytest.mark.parametrize(
    "grid_size, drone_amount",
    [
        (5, 25),
        (10, 100),
        (20, 400),
    ],
)
def test_maximum_drones_allowed(grid_size, drone_amount):
    try:
        env = init_drone_swarm_search(grid_size=grid_size, drone_amount=drone_amount)
        _ = env.reset()
    except ValueError as e:
        pytest.fail(
            f"The system should not fail with the maximum allowed number of drones. Error: {str(e)}"
        )

    assert (
        len(env.get_agents()) == drone_amount
    ), f"There should be {drone_amount} drones, but found {len(env.get_agents())}."


@pytest.mark.parametrize(
    "drone_amount",
    [
        2,
    ],
)
def test_drone_collision_termination(drone_amount):

    env = init_drone_swarm_search(drone_amount=drone_amount)
    _ = env.reset()

    done = False
    while not done:
        actions = {"drone0": Actions.SEARCH.value, "drone1": Actions.LEFT.value}
        _, reward, terminations, truncations, _ = env.step(actions)
        done = any(truncations.values()) or any(terminations.values())

        assert done, "The simulation should terminate upon drone collision."
        assert any(
            terminations.values()
        ), "There should be a termination flag set due to the collision."
        assert (
            sum(reward.values()) < 0
        ), "The total reward should be negative after a collision."


@pytest.mark.parametrize(
    "timestep_limit",
    [
        10,
        20,
        30,
        40,
        50,
    ],
)
def test_timeout_termination(timestep_limit):
    env = init_drone_swarm_search(timestep_limit=timestep_limit)
    _ = env.reset()

    done = False
    timestep_counter = 0
    while not done:
        actions = {"drone0": Actions.SEARCH.value}
        _, reward, terminations, done, _ = env.step(actions)
        done = any([e for e in done.values()])

        if timestep_counter > timestep_limit:
            assert (
                sum(reward.values()) < 0
            ), f"The total reward should be negative, indicating a penalty. But the total reward was: {reward['total_reward']}."
            assert (
                done
            ), "The simulation should flag as done after exceeding the timestep limit."
            assert any(
                terminations.values()
            ), "A termination flag should be set due to exceeding the timestep limit."
        timestep_counter += 1
    assert (
        timestep_counter > timestep_limit
    ), "The simulation should run beyond the timestep limit before terminating."


@pytest.mark.parametrize(
    "grid_size, person_initial_position",
    [
        (15, (4, 4)),
        (20, (10, 10)),
        (25, (15, 15)),
        (30, (20, 20)),
    ],
)
def test_leave_grid_get_negative_reward(grid_size, person_initial_position):
    env = init_drone_swarm_search(
        grid_size=grid_size, person_initial_position=person_initial_position
    )
    opt = {"drones_positions": [(0, 0)]}
    _ = env.reset(options=opt)

    done = False
    reward_sum = 0
    while not done and reward_sum >= env.reward_scheme.leave_grid * (env.timestep_limit - 1):
        actions = {"drone0": Actions.UP.value}
        _, reward, terminations, done, _ = env.step(actions)
        done = any(done.values())
        reward_sum += sum(reward.values())

    assert (
        not done
    ), "The simulation should not end, indicating the drone left the grid or another termination condition was met."
    assert (
        sum(reward.values()) < 0
    ), "The total reward should be negative, indicating a penalty was applied."
    assert not any(
        terminations.values()
    ), "There not should be at least one termination condition met."


@pytest.mark.parametrize(
    "grid_size, person_position",
    [
        (5, (19, 19)),
        (5, (-1, 0)),
        (5, (0, -1)),
        (5, (5, 5)),
    ],
)
def test_should_raise_invalid_person_position(grid_size, person_position):
    with pytest.raises(ValueError):
        init_drone_swarm_search(
            grid_size=grid_size, person_initial_position=person_position
        )


@pytest.mark.parametrize(
    "drone_amount",
    [
        5,
        10,
        15,
        20,
    ],
)
def test_if_all_drones_are_created(drone_amount):
    env = init_drone_swarm_search(drone_amount=drone_amount)
    _ = env.reset()

    assert (
        len(env.get_agents()) == drone_amount
    ), f"Should have {drone_amount} drones, but found {len(env.get_agents())}."


@pytest.mark.parametrize(
    "drone_amount, drones_positions",
    [
        (1, [(3, 3)]),
        (2, [(12, 0), (0, 13)]),
        (3, [(0, 0), (19, 19), (15, 10)]),
        (4, [(5, 0), (0, 0), (1, 1), (10, 10)]),
    ],
)
def test_position_drone_is_correct_after_reset(drone_amount, drones_positions):
    env = init_drone_swarm_search(drone_amount=drone_amount)

    opt = {"drones_positions": drones_positions}
    observations, _ = env.reset(options=opt)

    for i, position in enumerate(drones_positions):
        drone_id = f"drone{i}"
        real_position_drone = observations[drone_id][0]

        assert (
            real_position_drone == position
        ), f"Expected {drone_id}'s position to be {position} after reset, but was {real_position_drone}."


@pytest.mark.parametrize(
    "drone_amount, drones_positions",
    [
        (1, [(-1, 3)]),
        (2, [(12, 0), (25, 13)]),
        (3, [(0, 0), (19, 19), (25, -10)]),
        (4, [(5, 0), (0, 0), (10, 10), (10, 10)]),
    ],
)
def test_invalid_drone_position_raises_error(drone_amount, drones_positions):
    with pytest.raises(ValueError):
        env = init_drone_swarm_search(drone_amount=drone_amount)
        opt = {"drones_positions": drones_positions}
        _ = env.reset(options=opt)


@pytest.mark.parametrize(
    "drone_amount",
    [
        1,
        20,
        35,
        48,
    ],
)
def test_if_all_drones_are_created_with_default_positions(drone_amount):
    env = init_drone_swarm_search(drone_amount=drone_amount)

    _ = env.reset()

    assert (
        len(env.get_agents()) == drone_amount
    ), f"Should have {drone_amount} drones, but found {len(env.get_agents())}."


@pytest.mark.parametrize(
    "drone_amount, grid_size",
    [
        (1, 10),
        (2, 15),
        (5, 20),
        (15, 25),
    ],
)
def test_with_the_observation_size_is_correct_for_all_drones(drone_amount, grid_size):
    env = init_drone_swarm_search(grid_size=grid_size, drone_amount=drone_amount)

    observations, _ = env.reset()

    for drone in range(drone_amount):
        drone_id = f"drone{drone}"
        observation_matriz = observations[drone_id][1]

        assert observation_matriz.shape == (
            grid_size,
            grid_size,
        ), f"The observation matrix for {drone_id} should have a shape of ({grid_size}, {grid_size}), but was {observation_matriz.shape}."


def test_petting_zoo_interface_works():
    env = init_drone_swarm_search()
    parallel_api_test(env)
    env.close()


@pytest.mark.parametrize(
    "person_initial_position, person_amount",
    [
        ((10, 10), 10),
        ((10, 10), 15),
        ((10, 10), 20),
        ((10, 10), 25),
    ],
)
def test_castaway_count_after_reset(person_initial_position, person_amount):
    env = init_drone_swarm_search(
        person_amount=person_amount, person_initial_position=person_initial_position
    )
    _ = env.reset()

    assert (
        len(env.get_persons()) == person_amount
    ), f"Should have {person_amount} castaways, but found {len(env.get_persons())}."


@pytest.mark.parametrize(
    "person_initial_position, person_amount, drone_amount",
    [
        ((10, 10), 1, 1),
        ((1, 10), 5, 1),
        ((19, 5), 10, 1),
        ((5, 16), 15, 1),
    ],
)
def test_castaway_count_after_reset(
    person_initial_position, person_amount, drone_amount
):
    env = init_drone_swarm_search(
        person_amount=person_amount,
        person_initial_position=person_initial_position,
        drone_amount=drone_amount,
    )
    observations = env.reset()

    rewards = 0
    done = False
    while not done:
        actions = policy(observations, env.get_agents(), env)
        observations, reward, _, done, info = env.step(actions)
        rewards += sum(reward.values())
        done = any(done.values())

    _ = env.reset()

    assert (
        rewards >= DroneSwarmSearch.reward_scheme.search_and_find * person_amount
    ), f"The total reward should be positive after finding all castaways. But the total reward was: {rewards}."
    assert done, "The simulation should end after finding all castaways."
    assert (
        len(env.get_persons()) == person_amount
    ), f"Should have {person_amount} castaways, but found {len(env.get_persons())}."
    assert (
        len(env.get_agents()) == drone_amount
    ), f"Should have {drone_amount} drones, but found {len(env.get_agents())}."


@pytest.mark.parametrize(
    "pre_render_time, cell_size, drone_max_speed, wind_resistance",
    [
        (1, 130, 10, 0.0),
        (5, 130, 20, 0.0),
        (10, 130, 30, 0.0),
        (15, 130, 40, 0.0),
        (20, 130, 50, 0.0),
    ],
)
def test_pre_render_work_after_reset(
    pre_render_time, cell_size, drone_max_speed, wind_resistance
):
    env = init_drone_swarm_search(
        pre_render_time=pre_render_time, drone_speed=drone_max_speed
    )
    _ = env.reset()
    pre_render_steps = round(
        (pre_render_time * 60) / (cell_size / (drone_max_speed - wind_resistance))
    )

    assert (
        env.pre_render_steps == pre_render_steps
    ), f"The pre-render time should be {pre_render_steps}, but was {env.pre_render_time}."

    _ = env.reset()

    assert (
        env.pre_render_steps == pre_render_steps
    ), f"The pre-render time should be {pre_render_steps}, but was {env.pre_render_time}."


@pytest.mark.parametrize(
    "person_amount, mult",
    [
        (1, ["1"]),
        (2, [1, "0.8"]),
        (3, [1, "0.8", 0.7]),
        (4, ["1", 0.8, "0.7", 0.6]),
        (5, ["1", "0.8", "0.7", "0.6", "0.5"]),
    ],
)
def test_get_wrong_if_scale_pod_is_not_a_number(person_amount, mult):
    with pytest.raises(Exception):
        env = init_drone_swarm_search(person_amount=person_amount)
        opt = {"person_pod_multipliers": mult}
        _ = env.reset(options=opt)


@pytest.mark.parametrize(
    "person_amount, mult",
    [
        (1, [-1.2]),
        (2, [1, -0.8]),
        (3, [1, -0.8, 1.7]),
        (4, [1, 0.8, -0.7, 0.6]),
        (5, [1, 0.8, -0, 0.6, -3.5]),
    ],
)
def test_get_wrong_if_scale_mult_is_not_greater_than_0(person_amount, mult):
    with pytest.raises(Exception):
        env = init_drone_swarm_search(person_amount=person_amount)
        opt = {"person_pod_multipliers": mult}
        _ = env.reset(options=opt)


@pytest.mark.parametrize(
    "person_amount, mult",
    [
        (1, [1, 0.1]),
        (2, [1]),
        (3, [1, 0.8, 0.7, 1]),
        (4, [1, 0.8]),
        (5, [1, 0.8, 0.7, 0.6, 0.5, 0.6, 0.5]),
    ],
)
def test_get_wrong_if_number_of_mults_is_not_equal_to_person_amount(
    person_amount, mult
):
    with pytest.raises(Exception):
        env = init_drone_swarm_search(person_amount=person_amount)
        opt = {"person_pod_multipliers": mult}
        _ = env.reset(options=opt)



from dataclasses import dataclass


@dataclass
class DroneData:
    """
    Class to wrap the drone data.

    Attributes:
    -----------
    amount: int
        The number of drones in the environment.
    speed: float
        The speed of the drone in m/s.
    """

    amount: int
    speed: float
    pod: float = 1

    def __post_init__(self):
        if self.amount <= 0:
            raise ValueError("The number of drones must be greater than 0.")
        if self.speed <= 0:
            raise ValueError("The drone's speed must be greater than 0.")
        if self.pod < 0 or self.pod > 1:
            raise ValueError("The probability of detection must be between 0 and 1.")



import numpy as np
from numba import njit
from .time_step import calculate_time_step


class ProbabilityMatrix:
    """
    Probability matrix that represents the probability of finding a target in a certain position.

    Attributes
    ----------
    amplitude : int
        Amplitude of the gaussian function
    supposed_position : tuple
        Center of the gaussian function
    map, map_prob : np.array
        Matrix that represents the probability of finding a target in a certain position
    movement_vector : tuple
        Vector that determines the movement of the matrix
    inc_x : float
        Increment of the x position of the matrix
    inc_y : float
        Increment of the y position of the matrix
    spacement_inc : float
        Increment of the spacement of the gaussian function
    spacement : float
        Spacement of the gaussian function
    time_step_relation : int
        Relation between the matrix time step and the drone time step
    """

    def __init__(
        self,
        amplitude: int,
        spacement_start: float,
        spacement_inc: float,
        vector: tuple[float, float],
        initial_position: tuple[int, int],
        size: int,
    ):
        """
        Probability matrix that represents the probability of finding a target in a certain position

        Parameters
        ----------
        amplitude : int
            Amplitude of the gaussian function
        spacement_start : float
            Initial spacement of the gaussian function
        spacement_inc : float
            Increment of the spacement of the gaussian function
        vector : tuple
            Vector that determines the movement of the matrix
        initial_position : tuple
            Initial position of the matrix
        size : int
            Size of the matrix (size x size)
        """

        self.amplitude = amplitude
        self.supposed_position = initial_position
        self.map = np.zeros((size, size), dtype=np.float32)
        self.map_prob = np.zeros((size, size), dtype=np.float32)
        self.movement_vector = vector

        # These determine the movement of the target as well
        self.inc_x = 0
        self.inc_y = 0

        # These determine the shape of the gaussian function
        self.spacement_inc = spacement_inc
        self.spacement = spacement_start

        # Time step control
        self.time_step_relation = 1

    def step(self) -> None:
        self.update_position()
        self.diffuse_probability()

    def update_position(self) -> None:
        self.increment_movement()

        # On row, column notation (y, x)
        new_position = (
            self.supposed_position[0] + int(self.inc_y),
            self.supposed_position[1] + int(self.inc_x),
        )

        if self.is_valid_position(new_position):
            self.supposed_position = new_position

        if abs(self.inc_x) >= 1:
            self.inc_x -= np.sign(self.inc_x)
        if abs(self.inc_y) >= 1:
            self.inc_y -= np.sign(self.inc_y)

    def is_valid_position(self, position: tuple[int]) -> bool:
        rows, columns = self.map.shape
        is_valid_y = position[0] >= 0 and position[0] < rows
        is_valid_x = position[1] >= 0 and position[1] < columns
        return is_valid_x and is_valid_y

    def diffuse_probability(self) -> None:
        map_copy = self.calc_all_probs(
            self.amplitude,
            self.supposed_position,
            self.spacement,
            self.map.shape,
        )

        map_copy_sum = map_copy.sum()
        if map_copy_sum == 0:
            self.map_prob = map_copy
        else:
            self.map_prob = map_copy / map_copy_sum

        self.map = map_copy
        self.spacement += self.spacement_inc

    def increment_movement(self) -> None:
        self.inc_x += self.movement_vector[0] / self.time_step_relation
        self.inc_y += self.movement_vector[1] / self.time_step_relation

    @staticmethod
    @njit(cache=True, fastmath=True)
    def calc_all_probs(
        amplitude: int,
        supposed_position: tuple,
        spacement: float,
        shape: tuple,
    ) -> np.array:
        probabilities = np.zeros(shape, dtype=np.float32)
        # On row, column notation (y, x)
        x0 = supposed_position[1]
        y0 = supposed_position[0]
        for row in range(shape[0]):
            for column in range(shape[1]):
                x = ((column - x0) ** 2) / (2 * (spacement**2))
                y = ((row - y0) ** 2) / (2 * (spacement**2))
                probabilities[row][column] = amplitude * np.exp(-(x + y))
        return probabilities

    def update_time_step_relation(self, time_step: float, cell_size: float) -> None:
        self.time_step_relation = calculate_time_step(
            time_step, self.movement_vector, cell_size
        )

    def get_matrix(self):
        return self.map_prob
